#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "info.h"
#include "server.h"
#include "common.h"

#include <stdio.h>
#include <string.h>

#define IP_LEN		15	/* bf2 has IPv4 only */
#define NAME_LEN	80
#define PLAYERS_LEN	2
#define RTT_LEN		9	/* xxxx/xxxx */
#define GAME_LEN	8
#define MAP_LEN		80

#define LINE_LEN	255



struct print {
	struct server *s;

	char ip[IP_LEN + 1];
	char name[NAME_LEN + 1];
	char players[PLAYERS_LEN + 1];
	char rtt[RTT_LEN + 1];
	char game[GAME_LEN + 1];
	char map[MAP_LEN + 1];
};


struct field {
	size_t	offset;
	size_t	len;
};


struct fields {
	struct field ip;
	struct field name;
	struct field players;
	struct field rtt;
	struct field game;
	struct field map;
};



static void
build_stats(struct server *s)
{
	struct info *info = (struct info *) s->data;

	info_server_build(&info->info);
	rtt_make_stats(&info->rtt, s->reqs, s->req_n);
}



static void
prepare_server(struct print *p, struct settings *set)
{
	struct info *info = (struct info *) p->s->data;
	struct rtt *rtt = &info->rtt;
	int n;

	n = snprintf(p->ip, IP_LEN, "%s", p->s->addr);
	p->ip[n] = '\0';

	if (info->info.hostname != NULL) {
		n = snprintf(p->name, NAME_LEN, "%s",
				info->info.hostname);
	} else {
		n = 0;
	}
	p->name[n] = '\0';

	if (info->info.players == 0) {
		n = snprintf(p->players, PLAYERS_LEN, "-");
	} else {
		n = snprintf(p->players, PLAYERS_LEN, "%d",
				info->info.players);
	}
	p->players[n] = '\0';

	n = snprintf(p->rtt, RTT_LEN, "%d/%.3f", rtt->med, rtt->sdev);
	p->rtt[n] = '\0';

	n = snprintf(p->game, GAME_LEN, "%s", info->info.gamevariant);
	p->game[n] = '\0';

	n = snprintf(p->map, MAP_LEN, "%s", info->info.mapname);
	p->map[n] = '\0';

}



static void
print_field(char *line, struct field *field, char *str)
{
	int n;

	if (field->offset == -1) {
		return;
	}

	n = snprintf(line + field->offset, field->len + 1, "%s", str);
	line[field->offset + n] = ' ';
}



static void
print_server(struct fields *fields, struct print *print, struct settings *set)
{
	struct info *info;
	static char line[LINE_LEN + 1];
	int i;

	if (print->s != NULL) {
		info = (struct info *) print->s->data;
		
		if (print->s->fd == 0 || info->info.hostname == NULL) {
			snprintf(print->name, NAME_LEN, "-");
			snprintf(print->rtt, RTT_LEN, "-");
		}
		if (info->info.players == 0 && set->show_empty == 0) {
			return;
		}
	}

	if (set->cols > LINE_LEN) {
		set->cols = LINE_LEN;
	}

	memset(line, (int) ' ', set->cols);
	line[set->cols] = '\0';

	print_field(line, &fields->ip, print->ip);
	print_field(line, &fields->name, print->name);
	print_field(line, &fields->players, print->players);
	print_field(line, &fields->rtt, print->rtt);
	print_field(line, &fields->game, print->game);
	print_field(line, &fields->map, print->map);

	for (i = set->cols; line[i] == ' '; i--) {
		line[i] = '\0';
	}

	printf("%s\n", line);
}



static inline int
max_len(size_t old, char *str)
{
	size_t l = strlen(str);
	return (l > old) ? l : old;
}



static void
print_servers(struct print *p, size_t n, struct settings *set)
{
	size_t games_n = 0;
	size_t i, j;

	struct fields f;

	memset(&f, 0, sizeof(f));
	f.map.offset = -1;

	for (i = 0; i < n; i++) {
		int found = 0;

		f.ip.len = max_len(f.ip.len, p[i].ip);
		f.name.len = max_len(f.name.len, p[i].name);
		f.rtt.len = max_len(f.rtt.len, p[i].rtt);
		f.game.len = max_len(f.game.len, p[i].game);
		f.map.len = max_len(f.map.len, p[i].map);

		for (j = 0; j < i; j++) {
			if (strcmp(p[i].game, p[j].game) == 0) {
				found = 1;
			}
		}
		if (found == 0 && p->s != NULL) {
			games_n++;
		}
	}

	if (games_n < 2) {
		f.game.offset = -1;
		f.rtt.offset = set->cols - f.rtt.len;
	} else {
		f.game.offset = set->cols - f.game.len;
		f.rtt.offset = f.game.offset - 2 - f.rtt.len;
	}

	f.players.len = 2;
	f.players.offset = f.rtt.offset - 2 - f.players.len;

	f.name.offset = f.ip.offset + 2 + f.ip.len;
	f.name.len = f.players.offset - f.name.offset - 2;

	f.ip.offset = 0;

	for (i = 0; i < n; i++) {
		print_server(&f, &p[i], set);
	}
}



static void
print_server_verbose(struct server *s, struct settings *set)
{
	struct info *info = (struct info *) s->data;
	struct rtt *rtt = &info->rtt;

	if (s->fd == 0 || info->info.hostname == NULL) {
		snprintf(info->info.gamevariant, GAME_LEN, "-");
		snprintf(info->info.mapname, MAP_LEN, "-");
		snprintf(info->info.hostname, NAME_LEN, "-");
		return;
	}

	if (info->info.players == 0 && set->show_empty == 0) {
		return;
	}

	printf("%s/%s: [%s/%s] %s\n",
			s->addr, s->port,
			info->info.gamevariant,
			info->info.mapname,
			info->info.hostname != NULL ? info->info.hostname : "");
	printf("\t%d players\n", info->info.players);
	printf("\tRTT min/avg/med/max/sdev = %d/%d/%d/%d/%.3f ms, "
			"loss %.2f%% (%d/%d)\n",
			rtt->min, rtt->avg, rtt->med, rtt->max, rtt->sdev,
			(float) (s->req_n - s->req_r) / s->req_n * 100.0,
			s->req_n - s->req_r, s->req_n);
}



static void
print_servers_verbose(struct server *servers, size_t n, struct settings *set)
{
	size_t i;
	for (i = 0; i < n; i++) {
		build_stats(&servers[i]);
		print_server_verbose(&servers[i], set);
	}
}



void
info_print(struct server *servers, size_t n, struct settings *set)
{
	struct print *print;
	size_t i;

	if (set->info_level >= VERBOSE_VERBOSE) {
		print_servers_verbose(servers, n, set);
		return;
	}

	print = malloc(sizeof(struct print) * (n + 1));
	if (print == NULL) { /* TODO */
		return;
	}

	snprintf(print[0].ip, IP_LEN, "IP");
	snprintf(print[0].name, NAME_LEN, "Server");
	snprintf(print[0].players, PLAYERS_LEN, "Players");
	snprintf(print[0].rtt, RTT_LEN, "RTT");
	snprintf(print[0].game, GAME_LEN, "Game");
	snprintf(print[0].map, MAP_LEN, "Map");

	for (i = 1; i < n + 1; i++) {
		print[i].s = &servers[i - 1];
		build_stats(print[i].s);
		prepare_server(&print[i], set);
	}

	print_servers(print, n + 1, set);

	free(print);
}
