#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "common.h"
#include "info.h"
#include "network.h"

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#define DEFAULT_PORT	"29900"



/*
 * General packet format
 * -----------------------------------------------------------------------------
 *
 * Query
 * -------------------------------------
 * HEAD TT ---ID--- DATA
 *
 * Response
 * -------------------------------------
 * TT ---ID--- DATA
 *
 * Fields
 * -------------------------------------
 * 
 * HEAD is always fefd
 *
 * TT, packet types:
 *   00: status query
 *   09: ping
 *
 * ID, session id, 32-bit random number
 *
 *
 * Examples
 * -----------------------------------------------------------------------------
 *
 * Ping
 * -------------------------------------
 * query: fefd 09 2aea081f
 *        HEAD TT ---ID---
 *
 * response:   09 2aea081f 3000
 *             TT ---ID--- ????
 *
 * Query with multiple packet response
 * -------------------------------------
 * query: fefd 00 67458b6b 000000 01
 *        HEAD TT ---ID--- ?????? ??
 *
 * reply #0: 00 67458b6b 73706c69746e756d 00 0000
 *           TT ---ID--- s p l i t n u m  00  N C
 * reply #1: 00 67458b6b 73706c69746e756d 00 0101
 *           TT ---ID--- s p l i t n u m  00  N C
 * reply #2: 00 67458b6b 73706c69746e756d 00 8201
 *           TT ---ID--- s p l i t n u m  00  N C
 *
 * N: bin 00000000		f: final packet
 *        fsssssss		s: sequence number
 *
 * C: 00 for first packet, 01 for the others
 */



static void
logger(int level, const char *fmt, ...)
{
	va_list va;
	static char buf[BUF_LEN + 1];
	buf[BUF_LEN] = '\0';

	if (verbose < level) {
		return;
	}

	va_start(va, fmt);
	vsnprintf(buf, BUF_LEN, fmt, va);
	va_end(va);

	printf("%s", buf);
}



static void
make_status_query(uint32_t session, int server, int player, int team,
		char **data, size_t *len)
{
	static char t[] = {
		0xfe, 0xfd,		/* 0x00: header */
		0x00,			/* 0x02: packet type: status query */
		0x00, 0x00, 0x00, 0x00,	/* 0x03: session */
		0x00,			/* 0x07: server info and rules */
		0x00,			/* 0x08: player info */
		0x00,			/* 0x09: team info */
		0x01,			/* 0x0a: ??? */
	};
	uint32_t s;

	s = htonl(session);
	memcpy(t + 0x03, &s, 4);

	if (server == 1) {
		t[0x07] = 0xff;
	}
	if (player == 1) {
		t[0x08] = 0xff;
	}
	if (team == 1) {
		t[0x09] = 0xff;
	}

	*data = t;
	*len = sizeof(t);
}



static void
make_ping(uint32_t session, char **data, size_t *len)
{
	static char t[] = {
		0xfe, 0xfd,		/* 0x00: header */
		0x09,			/* 0x02: packet type: ping */
		0x00, 0x00, 0x00, 0x00	/* 0x03: session id */
	};
	uint32_t s;

	s = htonl(session);
	memcpy(t + 0x03, &s, 4);
	*data = t;
	*len = sizeof(t);
}



static void
hexdump(unsigned char *data, size_t len)
{
	char out[68];
	char tmp[3];
	size_t i, n;

	n = 0;

	for (i = 0; i < len; i++) {
		int space = 6 + n / 4 + n / 8;
		if (n == 0) {
			memset(out, (int) ' ', sizeof(out));
			out[66] = '\0';
			sprintf(out, "%04x: ", (unsigned int) i);
			out[6] = ' ';
		}

		sprintf(tmp, "%02x", data[i]);
		memcpy(out + n * 2 + space, tmp, 2);
		switch (data[i]) {
		case 0x20 ... 0x7e:
			out[40 + n + space] = data[i];
			break;
		default:
			out[40 + n + space] = '.';
			break;
		}

		n++;
		if (n == 16) {
			printf("%s\n", out);
			n = 0;
		}
	}

	if (n != 0) {
		printf("%s\n", out);
	}
}



static void
add_request(struct server *s, enum packet type)
{
	struct request *req;

	s->req_n++;
	s->reqs = realloc(s->reqs, sizeof(struct request) * s->req_n);
	assert(s->reqs != NULL);
	req = &s->reqs[s->req_n - 1];
	memset(req, (int) 0, sizeof(struct request));

	req->session = rand() & 0xffffffff;
	req->sent.tv_sec = req->sent.tv_usec = 0;
	req->recv.tv_sec = req->recv.tv_usec = 0;
	req->type = type;

	if (verbose >= VERBOSE_DEBUG) {
		printf("server %p: added request %p: "
				"type=%d session=%x (count=%d)\n",
				(void *) s, (void *) req,
				(int) req->type,
				(unsigned int) req->session,
				s->req_n);
	}
}



#if 0
static void
check_writeable(struct server *servers, size_t n, fd_set *fds)
{
	size_t i;
	
	for (i = 0; i < n; i++) {
		struct server *s = &servers[i];
		int r;
		int err;
		socklen_t errlen;

		if (s->fd == 0 || ! FD_ISSET(s->fd, fds)) {
			continue;
		}

		errlen = sizeof(err);
		r = getsockopt(s->fd, SOL_SOCKET, SO_ERROR, &err, &errlen);
		if (r == -1) {
			server_error(s, 1, "Reading connect result failed",
					strerror(errno));
			continue;
		}
		if (err != 0) {
			server_error(s, 1, "Cannot connect", strerror(err));
			continue;
		}

		s->ci.write = 0;
		s->ci.read = 1;
	}
}
#endif



static void
send_server_packet(struct server *s)
{
	struct request *req;
	char *data;
	size_t len;
	ssize_t r;

	if (s->fd == 0 || s->ci.read == 0 || s->req_i == s->req_n) {
		return;
	}

	req = &s->reqs[s->req_i];
	s->req_i++;

	switch (req->type) {
		case PACKET_PING:
			make_ping(req->session, &data, &len);
			break;
		case PACKET_INFO:
			make_status_query(req->session, 0, 1, 0,
					&data, &len);
			break;
		default:
			assert(0);
	}

	if (len == 0) {
		return;
	}

	if (verbose >= VERBOSE_FLOOD) {
		printf("\n\nsend %d bytes\n", (int) len);
		hexdump((unsigned char *) data, len);
	}
	gettimeofday(&req->sent, NULL);
	r = send(s->fd, data, len, 0);
	if (r != len) {
		server_error(s, 1, "Sending data failed",
				strerror(errno));
	}
}



static void
recv_pong(struct server *s, struct request *req, struct timeval *now)
{
	if (verbose >= VERBOSE_VERBOSE) {
		int rtt = timediff(now, &req->sent);
		printf("%s/%s: Pong %d ms\n", s->addr, s->port, rtt);
	}
	s->req_r++;
}



static void
recv_info(struct server *s, struct request *req, struct timeval *now,
		const char *buf, size_t len, struct settings *set)
{
	struct info *info;
	int s_final = buf[14] & 0x80;
/*	int s_no = buf[14] & 0x7f; */
	int s_id = buf[15];

	if (len < 16) {
		server_error(s, 1, NULL, "Bad info packet reply");
		return;
	}

	if (verbose >= VERBOSE_VERBOSE) {
		printf("%s/%s: Info packet %d%s\n",
				s->addr, s->port,
				s_id, s_final == 0 ? "" : ", final");
	}

	info = (struct info *) s->data;
	memcpy(info->info.data + info->info.ptr, buf + 16, len - 16);
	info->info.ptr += len - 16;

	if (s_final != 0) {
		size_t i;
		s->req_r++;
		for (i = 0; i < set->ping; i++) {
			add_request(s, PACKET_PING);
		}
	}
}



static void
check_readable(struct server *servers, size_t n, fd_set *fds,
		struct settings *set)
{
	char buf[BUF_LEN + 1];
	size_t i;
	int j;

	buf[BUF_LEN] = '\0';

	for (i = 0; i < n; i++) {
		struct server *s = &servers[i];
		struct timeval now;
		int r;
		int type;
		uint32_t session;

		if (s->fd == 0) {
			continue;
		}

		gettimeofday(&now, NULL);
		if (!FD_ISSET(s->fd, fds)) {
			if (timediff(&now, &s->reqs[s->req_i - 1].sent) >=
					1000) {
				send_server_packet(s);
			}
			continue;
		}

		r = recv(s->fd, buf, BUF_LEN, 0);
		if (r <= 0) {
			if (errno != ECONNREFUSED ||
					verbose >= VERBOSE_VERBOSE) {
				server_error(s, 1, "Receiving data failed",
						strerror(errno));
			}
			continue;
		}

		if (verbose >= VERBOSE_FLOOD) {
			printf("\nreceived %d bytes\n", r);
			hexdump((unsigned char *) buf, r);
		}

		type = (int) *((unsigned char *) buf);
		session = ntohl(*((uint32_t *) (buf + 1)));

		for (j = 0; j < s->req_n; j++) {
			struct request *req = &s->reqs[j];
			if (req->session != session) {
				continue;
			}
			gettimeofday(&req->recv, NULL);
			
			switch (type) {
				case PACKET_PING:
					recv_pong(s, req, &now);
					break;
				case PACKET_INFO:
					recv_info(s, req, &now, buf, r, set);
					break;
				default:
					break;
			}
		}

		send_server_packet(s);
	}
}



static void
send_first_packet(struct server *servers, size_t n)
{
	size_t i;
	for (i = 0; i < n; i++) {
		struct server *s = &servers[i];
		send_server_packet(s);
	}
}



static int
query_finished(struct server *servers, size_t n, int timeout)
{
	struct timeval now;
	size_t i;
	int to = 0;

	gettimeofday(&now, NULL);

	for (i = 0; i < n; i++) {
		struct server *s = &servers[i];
		struct request *req;
		
		if (s->fd == 0) {
			logger(VERBOSE_DEBUG, "Timeout=%d (no FD)\n", to + 1);
			to++;
			continue;
		}
		if (s->req_i == 0) {
			continue;
		}

		if (s->req_r == s->req_n) {
			to++;
			continue;
		}

		req = &s->reqs[s->req_i - 1];
		
		if (req->recv.tv_sec == 0) {
			if (req->sent.tv_sec != 0 &&
					timediff(&now, &req->sent) > timeout) {
				to++;
				logger(VERBOSE_DEBUG,
						"Timeout for first reply: "
						"to=%d\n", to);
			}
		} else if (timediff(&now, &req->recv) > timeout) {
			to++;
			logger(VERBOSE_DEBUG, "Timeout after recv: to=%d\n", to);
		}
	}

	return to == n ? 1 : 0;
}



static void
retry_info(struct server *servers, size_t n, int timeout)
{
	size_t i;
	struct timeval now;
	gettimeofday(&now, NULL);

	for (i = 0; i < n; i++) {
		struct server *s = &servers[i];
		struct request *req = &s->reqs[0];
		if (s->fd == 0) {
			continue;
		}
		if (req->recv.tv_sec == 0 && timediff(&now, &req->sent) >=
				timeout / 2) {
			logger(VERBOSE_VERBOSE,
					"%s/%s: Retrying to get server info\n",
					s->addr, s->port);
			add_request(s, PACKET_INFO);
		}
	}
}



static int
mainloop(struct server *servers, size_t n, struct settings *settings)
{
	fd_set rfds, wfds;
	struct timeval to;
	int r;

	while (query_finished(servers, n, settings->timeout) == 0) {
		int high;
		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		to.tv_sec = 0;
		to.tv_usec = 1000 * 1000;

		retry_info(servers, n, settings->timeout);

		high = network_set_fds(servers, n, &rfds, &wfds);
		r = select(high + 1, &rfds, &wfds, NULL, &to);
		if (r < 0) {
			fprintf(stderr, "select() failed: %s\n",
					strerror(errno));
			return -1;
		} else if (r == 0) {
			continue;
		}

#if 0
		check_writeable(servers, n, &wfds);
#endif
		check_readable(servers, n, &rfds, settings);
	}

	return 0;
}



static struct server *
parse_cmdline(int argc, char **argv, size_t *server_n, struct settings *set)
{
	struct server *servers;
	char *port;
	int i;

	set->show_players = 0;
	set->show_empty = 0;

	set->timeout = 5000;
	set->ping = 1;
	set->cols = 80; /* TODO */
	set->info_level = VERBOSE_BRIEF;

	while (1) {
		int opt;
		opt = getopt(argc, argv, "vp:lt:ei");
		if (opt == -1) {
			break;
		}

		switch (opt) {
			case 'v':
				verbose++;
				break;
			case 'l':
				set->show_players = 1;
				break;
			case 'p':
				set->ping = atoi(optarg);
				break;
			case 't':
				set->timeout = atoi(optarg);
				break;
			case 'e':
				set->show_empty = 1;
				break;
			case 'i':
				set->info_level++;
				break;
			case '?':
				fprintf(stderr, "? SYNTAX ERROR\n");
				exit(EXIT_FAILURE);
		}
	}

	*server_n = argc - optind;
	servers = calloc(sizeof(struct server), *server_n);
	assert(servers != NULL);

	for (i = optind; i < argc; i++) {
		struct server *s = &servers[i - optind];

		s->data = malloc(sizeof(struct info));
		if (s->data == NULL) {
			exit(EXIT_FAILURE); /* FIXME */
		}

		port = strchr(argv[i], (int) '/');
		if (port != NULL) {
			strncpy(s->port, port + 1, SERV_LEN);
			*port = '\0';
		} else {
			strncpy(s->port, DEFAULT_PORT, SERV_LEN);
		}
		strncpy(s->addr, argv[i], ADDR_LEN);

		/* set requests */
#if 0
		for (j = 0; j < set->ping; j++) {
			add_request(s, PACKET_PING);
		}
#endif
		add_request(s, PACKET_INFO);
	}

	return servers;
}



static void
free_servers(struct server *servers, size_t n)
{
	size_t i;
	for (i = 0; i < n; i++) {
		struct server *s = &servers[i];
		if (s->fd != 0) {
			close(s->fd);
		}
		free(s->data);
		free(s->reqs);
		freeaddrinfo(s->ci.res);
	}
	free(servers);
}



int
main(int argc, char **argv)
{
	struct settings settings;
	struct server *servers;
	size_t n;
	int r;

	servers = parse_cmdline(argc, argv, &n, &settings);
	if (servers == NULL) {
		return EXIT_SUCCESS;
	}
	network_connect_all(servers, n);
	send_first_packet(servers, n);
	r = mainloop(servers, n, &settings);
	info_print(servers, n, &settings);
	free_servers(servers, n);

	return r == 0 ? EXIT_SUCCESS : EXIT_FAILURE;
}
