#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "server.h"

#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>



static int
init_socket(struct server *s, int domain, int type, int protocol)
{
	int r, t;
#if 0
	int flags;
#endif

	s->fd = socket(domain, type, protocol);
	if (s->fd < 0) {
		server_error(s, 0, "Creating socket failed",
				strerror(errno));
		return -1;
	}

	/* reuse */
	r = setsockopt(s->fd, SOL_SOCKET, SO_REUSEADDR, (void *) &t, sizeof(t));
	if (r < 0) {
		server_error(s, 1, "Setting reuse address failed",
				strerror(errno));
		return -1;
	}

#if 0
	/* non-blocking */
	flags = fcntl(s->fd, F_GETFL, 0);
	if (flags == -1) {
		server_error(s, 1, "Setting socket non-blocking failed",
				strerror(errno));
		return -1;
	}
	flags |= O_NONBLOCK;
	r = fcntl(s->fd, F_SETFL, flags);
	if (r == -1) {
		server_error(s, 1, "Setting socket non-blocking failed",
				strerror(errno));
		return -1;
	}
#endif

	return 0;
}



static int
bind_socket(int s, const char *haddr, uint16_t port)
{
	struct hostent *info;
	struct sockaddr_in addr;
	int r;

	info = gethostbyname(haddr);
	if (info == NULL) {
		perror("gethostbyname(local)");
		return -1;
	}
	memcpy(&addr.sin_addr, info->h_addr, info->h_length);
	addr.sin_family = info->h_addrtype;
	if (port != 0) {
		addr.sin_port = htons(port);
	}

	r = bind(s, (struct sockaddr *) &addr, sizeof(addr));
	if (r < 0) {
		perror("bind()");
		return -1;
	}

	return 0;
}



static int
connect_single(struct server *s)
{
	int r;

	s->ci.hints.ai_family = PF_UNSPEC;
	s->ci.hints.ai_socktype = SOCK_DGRAM;
	s->ci.hints.ai_protocol = IPPROTO_UDP;

	r = getaddrinfo(s->addr, s->port, &s->ci.hints, &s->ci.res);
	if (r != 0) {
		server_error(s, 0, "Resolving address failed",
				gai_strerror(r));
		return 1;
	}

	r = init_socket(s, s->ci.res->ai_family,
			s->ci.res->ai_socktype,
			s->ci.res->ai_protocol);
	if (r == -1) {
		return 1;
	}

	if (0) bind_socket(s->fd, NULL, 0);

	r = connect(s->fd, s->ci.res->ai_addr, s->ci.res->ai_addrlen);
	if (r != 0) {
		server_error(s, 1, "Cannot connect", strerror(errno));
		return 1;
	} else {
		s->ci.read = 1;
		return 0;
	}
}



int
network_connect_all(struct server *servers, size_t n)
{
	size_t i;
	int done = 0;
	for (i = 0; i < n; i++) {
		done += connect_single(&servers[i]);
	}
	return done;
}



int
network_set_fds(struct server *servers, size_t n, fd_set *r, fd_set *w)
{
	size_t i;
	int high = 0;
	
	for (i = 0; i < n; i++) {
		struct server *s = &servers[i];
		if (s->fd != 0) {
			if (s->ci.read == 1) {
				FD_SET(s->fd, r);
				if (s->fd > high) {
					high = s->fd;
				}
			}
			if (s->ci.write == 1) {
				FD_SET(s->fd, w);
				if (s->fd > high) {
					high = s->fd;
				}
			}
		}
	}
	return high;
}
